from enum import Enum
from typing import Union

from aiogram import types
from aiogram.filters import Filter

from .asyncio import BaseAsyncStorage
from .storage import BaseStorage


class IsState(Filter):
    """
    A filter to validate the current state of a user using aiostep's state management.

    This filter checks if the user's current state matches the given `state`. It uses the `state_manager`
    to retrieve the user's state and validates it against the provided `state`.

    Args:
        state (str | Enum): The target state to match. If an `Enum` is provided, its `name` will be used.
        state_manager (BaseStorage | BaseAsyncStorage):
            An instance of aiostep's asynchronous state manager for retrieving and validating user states.

    Methods:
        __call__(event): Processes an input event and validates the user's current state.
            - event (types.Message | types.CallbackQuery): The input event to validate.
            - Returns (bool): True if the user's state matches the target state, False otherwise.
    """

    def __init__(
        self,
        state: Union[str, Enum],
        state_manager: Union[BaseStorage, BaseAsyncStorage],
    ) -> None:
        self.state = state.name if isinstance(state, Enum) else state
        self.state_manager = state_manager
        if isinstance(state_manager, BaseStorage):
            self.sync = True
        else:
            self.sync = False

    async def __call__(self, event: Union[types.Message, types.CallbackQuery]) -> bool:
        if self.sync:
            current_state = self.state_manager.get_state(event.from_user.id)
        else:
            current_state = await self.state_manager.get_state(event.from_user.id)
        if not current_state:
            return False

        chat_id = (
            event.message.chat.id
            if isinstance(event, types.CallbackQuery)
            else event.chat.id
        )

        return (current_state.current_state == self.state) and (
            current_state.chat_id == chat_id
        )



import asyncio
import typing

try:
    from telethon import TelegramClient, events

    telethon_installed = True
except ImportError:
    telethon_installed = False

from ..functions import MetaStore, root


def Listen(
    app: "TelegramClient",
    store: typing.Optional[MetaStore] = None,
    event: typing.Any = None,
    **kwargs,
) -> None:
    """
    telethon listen client for steps.

    supported events:
        - `NewMessage`
        - `CallbackQuery`
        - `MessageEdited`

    Example::

        app = TelegramClient()
        aiostep.telethon_dialect.Listen(app)
    """
    if not telethon_installed:
        raise ImportError(
            "telethon package is not installed. install package: pip install telethon"
        )
    store = store or root
    event = event or events.NewMessage

    async def _listen_wrapper(_event):
        fn = None

        try:
            fn = await store.pop_item(_event.sender_id)
        except (KeyError, AttributeError):
            try:
                fn = await store.pop_item(_event.chat_id)
            except (KeyError, AttributeError):
                pass

        if fn is not None:
            if isinstance(fn, asyncio.Future):
                fn.set_result(_event)
                return

            await fn(_event)
            return

    app.add_event_handler(_listen_wrapper, event(**kwargs))



import asyncio
import typing

try:
    from telebot import SkipHandler, types
    from telebot.async_telebot import BaseMiddleware

    telebot_installed = True
except ImportError:

    class BaseMiddleware:
        pass

    telebot_installed = False

from ..functions import MetaStore, root


class Listen(BaseMiddleware):
    """
    telebot middleware to listen for steps.

    supported update types:
        - `message`
        - `callback_query`
        - `chat_join_request`
        - `chat_member`
        - `edited_message`

    Example::

        bot = TeleBot()
        bot.setup_middleware(Listen(update_types=["message", "callback_query"]))
    """

    def __init__(
        self,
        update_types: typing.Optional[typing.List[str]] = None,
        store: typing.Optional[MetaStore] = None,
    ):
        if not telebot_installed:
            raise ImportError(
                "pyTelegramBotAPI package is not installed. "
                "install package: "
                "pip install pyTelegramBotAPI"
            )
        self.store = store or root
        self.update_sensitive = True
        self.update_types = update_types or ["message"]

    async def pre_process_message(self, message: "types.Message", data):
        fn = None

        try:
            fn = await self.store.pop_item(message.from_user.id)
        except (KeyError, AttributeError):
            try:
                fn = await self.store.pop_item(message.chat.id)
            except (KeyError, AttributeError):
                pass

        if fn is not None:
            if isinstance(fn, asyncio.Future):
                fn.set_result(message)
                return SkipHandler()
            await fn(message)
            return SkipHandler()

    async def post_process_message(self, message: "types.Message", data, exception):
        pass

    async def pre_process_callback_query(self, call: "types.CallbackQuery", data):
        fn = None

        try:
            fn = await self.store.pop_item(call.from_user.id)
        except (KeyError, AttributeError):
            try:
                fn = await self.store.pop_item(call.message.chat.id)
            except (KeyError, AttributeError):
                pass

        if fn is not None:
            if isinstance(fn, asyncio.Future):
                fn.set_result(call)
                return SkipHandler()
            await fn(call)
            return SkipHandler()

    async def post_process_callback_query(
        self, call: "types.CallbackQuery", data, exception
    ):
        pass

    async def pre_process_edited_message(self, message: "types.Message", data):
        fn = None

        try:
            fn = await self.store.pop_item(message.from_user.id)
        except (KeyError, AttributeError):
            try:
                fn = await self.store.pop_item(message.chat.id)
            except (KeyError, AttributeError):
                pass

        if fn is not None:
            if isinstance(fn, asyncio.Future):
                fn.set_result(message)
                return SkipHandler()
            await fn(message)
            return SkipHandler()

    async def post_process_edited_message(
        self, message: "types.Message", data, exception
    ):
        pass

    async def pre_process_chat_join_request(
        self, join_request: "types.ChatJoinRequest", data
    ):
        fn = None

        try:
            fn = await self.store.pop_item(join_request.from_user.id)
        except (KeyError, AttributeError):
            try:
                fn = await self.store.pop_item(join_request.chat.id)
            except (KeyError, AttributeError):
                pass

        if fn is not None:
            if isinstance(fn, asyncio.Future):
                fn.set_result(join_request)
                return SkipHandler()
            await fn(join_request)
            return SkipHandler()

    async def post_process_chat_join_request(
        self, join_request: "types.ChatJoinRequest", data, exception
    ):
        pass

    async def pre_process_chat_member(self, status: "types.ChatMemberUpdated", data):
        fn = None

        try:
            fn = await self.store.pop_item(status.from_user.id)
        except (KeyError, AttributeError):
            try:
                fn = await self.store.pop_item(status.chat.id)
            except (KeyError, AttributeError):
                pass

        if fn is not None:
            if isinstance(fn, asyncio.Future):
                fn.set_result(status)
                return SkipHandler()
            await fn(status)
            return SkipHandler()

    async def post_process_chat_member(
        self, status: "types.ChatMemberUpdated", data, exception
    ):
        pass



from enum import Enum
from typing import Any, Callable, Dict, Optional, Union

from msgspec import DecodeError
from msgspec.json import Decoder, Encoder

try:
    from redis.asyncio.client import Redis
    from redis.typing import ExpiryT

    redis_installed = True
except ImportError:
    redis_installed = False

from typing import Any, Callable

from ..storage.base import StateContext
from .base import BaseAsyncStorage


class AsyncRedisStateStorage(BaseAsyncStorage):
    """Redis-based storage implementation for managing bot states.

    This class provides a Redis-backed storage solution for managing bot states
    and their associated data. Suitable for production use and distributed systems.

    Note:
        Requires redis package to be installed:
        pip install your-package[redis]

    Args:
        cache (Redis): Redis client instance
        ex (ExpiryT | None): Optional expiration time for all keys
    """

    def __init__(
        self,
        redis: Optional["Redis"] = None,
        host: Optional[str] = "localhost",
        port: Optional[int] = 6379,
        db: Optional[int] = 0,
        password: Optional[str] = None,
        ex: Optional["ExpiryT"] = None,
        **kwargs,
    ) -> None:
        """Initialize the Redis storage.

        Args:
            cache (Redis): Redis client instance
            ex (ExpiryT | None, optional): Expiration time for all keys.
                Defaults to None.
        """
        if not redis_installed:
            raise ImportError(
                "Redis package is not installed. "
                "To use RedisStateStorage, install package with redis support: "
                "pip install aiostep[redis]"
            )

        if not redis:
            redis = Redis(host=host, port=port, db=db, password=password, **kwargs)
        self.cache = redis
        self.ex = ex
        self.encoder = Encoder()
        self.decoder = Decoder()

    def _get_key(self, user_id: Union[int, str]) -> str:
        """Generate Redis key for a user.

        Args:
            user_id (int | str): ID of the user

        Returns:
            str: Redis key
        """
        return f"state:{user_id}"

    def _get_data_key(self, user_id: Union[int, str]) -> str:
        """Generate Redis key for a user.

        Args:
            user_id (int | str): ID of the user

        Returns:
            str: Redis key
        """
        return f"data:{user_id}"

    async def set_state(
        self,
        user_id: Union[int, str],
        state: Union[str, Enum],
        callback: Optional[Callable[..., Any]] = None,
        chat_id: Optional[Union[int, str]] = None,
        ex: Optional["ExpiryT"] = None,
    ) -> None:
        """Set the state for a user.

        Args:
            user_id (int | str): ID of the user
            state (str | Enum): State to set
            callback (Callable | None, optional): Callback function. Defaults to None.
            chat_id (int | str, optional): Chat ID. Defaults to None.
        """
        if chat_id is None:
            chat_id = user_id

        if isinstance(state, Enum):
            state = state.name

        callback_name = callback.__name__ if callback else None

        state_data = {
            "current_state": state,
            "chat_id": chat_id,
            "callback": callback_name,
        }

        await self.cache.set(
            self._get_key(user_id), self.encoder.encode(state_data), ex=ex or self.ex
        )

    async def get_state(
        self, user_id: Union[int, str], default: Optional[Any] = None
    ) -> Optional[StateContext]:
        """Get the state context for a user.

        Args:
            user_id (int | str): ID of the user
            default (Any, optional): Default value if state doesn't exist.
                Defaults to None.

        Returns:
            StateContext | None: The state context or default value
        """
        data = await self.cache.get(self._get_key(user_id))
        if not data:
            return default

        state_data = self.decoder.decode(data)
        return StateContext(**state_data)

    async def delete_state(
        self, user_id: Union[int, str], default: Optional[Any] = None
    ) -> Optional[StateContext]:
        """Delete the state for a user.

        Args:
            user_id (int | str): ID of the user
            default (Any, optional): Default value if state doesn't exist.
                Defaults to None.

        Returns:
            StateContext | None: The deleted state context or default value
        """
        state_key = self._get_key(user_id)
        data = await self.cache.get(state_key)
        await self.cache.delete(state_key)

        if not data:
            return default

        state_data = self.decoder.decode(data)
        return StateContext(**state_data)

    async def set_data(
        self,
        user_id: Union[int, str],
        data: Optional[Dict[Any, Any]] = None,
        ex: Optional["ExpiryT"] = None,
        **kwargs,
    ) -> None:
        """Set data for a user's state.

        This method completely replaces any existing data.

        Args:
            user_id (int | str): ID of the user
            data (dict[str, Any]): Data to store
        """
        if data is not None and not isinstance(data, dict):
            raise ValueError(f"'data' must be a dict, got {type(data)}")

        data_payload = {}
        if data:
            data_payload.update(data)
        if kwargs:
            data_payload.update(kwargs)

        if not data_payload:
            raise ValueError("No data passed.")

        data_key = self._get_data_key(user_id)

        await self.cache.set(
            data_key, self.encoder.encode(data_payload), ex=ex or self.ex
        )

    async def get_data(
        self, user_id: Union[int, str], default: Optional[Any] = None
    ) -> Optional[Dict[Any, Any]]:
        """Get data for a user's state.

        Args:
            user_id (int | str): ID of the user
            default (Any, optional): Default value if data doesn't exist.
                Defaults to None.

        Returns:
            dict[str, Any] | None: The stored data or None if not found
        """
        data = await self.cache.get(self._get_data_key(user_id))
        if not data:
            return default

        state_data = self.decoder.decode(data)
        return state_data

    async def update_data(
        self,
        user_id: Union[int, str],
        data: Optional[Dict[Any, Any]] = None,
        ex: Optional["ExpiryT"] = None,
        **kwargs,
    ) -> None:
        """Update data for a user's state.

        This method updates existing data with new values, similar to dict.update().
        Existing keys will be updated, and new keys will be added.

        Args:
            user_id (int | str): ID of the user
            data (dict[str, Any]): Data to update

        Example:
            >>> # Existing data: {"name": "John"}
            >>> await storage.update_data(user_id, {"age": 25})
            >>> # Result: {"name": "John", "age": 25}
        """
        if data is not None and not isinstance(data, dict):
            raise ValueError(f"'data' must be a dict, got {type(data)}")

        data_payload = {}
        if data:
            data_payload.update(data)
        if kwargs:
            data_payload.update(kwargs)

        if not data_payload:
            raise ValueError("No data passed.")

        data_key = self._get_data_key(user_id)
        current_data = await self.cache.get(data_key)

        if current_data:
            try:
                state_data = self.decoder.decode(current_data)
                state_data.update(data_payload)
            except DecodeError:
                state_data = data_payload
        else:
            state_data = data_payload

        await self.cache.set(
            data_key, self.encoder.encode(state_data), ex=ex or self.ex
        )

    async def delete_data(
        self, user_id: Union[int, str], default: Optional[Any] = None
    ) -> Optional[Dict[Any, Any]]:
        """Clear and get all data for a user's state.

        Args:
            user_id (int | str): ID of the user
            default (Any, optional): Default value if data doesn't exist.
                Defaults to None.

        Returns:
            Dict | None: The deleted data or default value
        """
        data_key = self._get_data_key(user_id)
        data = await self.cache.get(data_key)
        await self.cache.delete(data_key)

        if not data:
            return default

        data = self.decoder.decode(data)
        return data



from copy import deepcopy
from enum import Enum
from typing import Any, Callable, Dict, Optional, Union

from cachebox import BaseCacheImpl, Cache, TTLCache

from ..storage.base import StateContext
from .base import BaseAsyncStorage


class AsyncMemoryStateStorage(BaseAsyncStorage):
    """In-memory storage implementation for managing bot states.

    This class provides an in-memory storage solution for managing bot states
    and their associated data. Suitable for development or small applications
    where persistence isn't required.

    Args:
        cache (dict | None): Optional dictionary to use as storage. If None,
            an empty dictionary will be used.
    """

    def __init__(
        self,
        cache: Optional[Union[BaseCacheImpl, dict]] = None,
        ex: Optional[float] = None,
    ) -> None:
        """Initialize the memory storage.

        Args:
            cache (dict | None, optional): Initial cache dictionary. Defaults to None.
            ex (float | None): expiry time for objects. pass in seconds.
        """
        if cache:
            self.cache = cache
        elif ex:
            self.cache = TTLCache(0, ex)
        else:
            self.cache = Cache(0)

    def _get_key(self, user_id: Union[int, str]) -> str:
        """Generate Cache key for a user.

        Args:
            user_id (int | str): ID of the user

        Returns:
            str: Cache key
        """
        return f"state:{user_id}"

    def _get_data_key(self, user_id: Union[int, str]) -> str:
        """Generate Cache key for a user.

        Args:
            user_id (int | str): ID of the user

        Returns:
            str: Cache key
        """
        return f"data:{user_id}"

    async def set_state(
        self,
        user_id: Union[int, str],
        state: Union[str, Enum],
        callback: Optional[Callable[..., Any]] = None,
        chat_id: Optional[Union[int, str]] = None,
    ) -> None:
        """Set the state for a user.

        Args:
            user_id (int | str): ID of the user
            state (str | Enum): State to set
            callback (Callable | None, optional): Callback function. Defaults to None.
            chat_id (int | str, optional): Chat ID. Defaults to None.
        """
        if chat_id is None:
            chat_id = user_id

        if isinstance(state, Enum):
            state = state.name

        state_key = self._get_key(user_id)

        self.cache[state_key] = StateContext(
            current_state=state, callback=callback, chat_id=chat_id
        )

    async def get_state(
        self, user_id: Union[int, str], default: Optional[Any] = None
    ) -> Optional[StateContext]:
        """Get the state context for a user.

        Args:
            user_id (int | str): ID of the user
            default (Any, optional): Default value if state doesn't exist.
                Defaults to None.

        Returns:
            StateContext | None: The state context or default value
        """
        state_key = self._get_key(user_id)
        return self.cache.get(state_key, default)

    async def delete_state(
        self, user_id: Union[int, str], default: Optional[Any] = None
    ) -> Optional[StateContext]:
        """Delete the state for a user.

        Args:
            user_id (int | str): ID of the user
            default (Any, optional): Default value if state doesn't exist.
                Defaults to None.

        Returns:
            StateContext | None: The deleted state context or default value
        """
        state_key = self._get_key(user_id)
        return self.cache.pop(state_key, default)

    async def set_data(
        self, user_id: Union[int, str], data: Optional[Dict[Any, Any]] = None, **kwargs
    ) -> None:
        """Set data for a user.

        This method completely replaces any existing data.

        Args:
            user_id (int | str): ID of the user
            data (dict[str, Any]): Data to store
        """
        if data is not None and not isinstance(data, dict):
            raise ValueError(f"'data' must be a dict, got {type(data)}")

        data_payload = {}
        if data:
            data_payload.update(data)
        if kwargs:
            data_payload.update(kwargs)

        if not data_payload:
            raise ValueError("No data passed.")

        data_key = self._get_data_key(user_id)

        self.cache[data_key] = deepcopy(data_payload)

    async def get_data(
        self, user_id: Union[int, str], default: Optional[Any] = None
    ) -> Optional[Dict[Any, Any]]:
        """Get data for a user.

        Args:
            user_id (int | str): ID of the user
            default (Any, optional): Default value if data doesn't exist.
                Defaults to None.

        Returns:
            dict[str, Any] | None: The stored data or None if not found
        """
        data_key = self._get_data_key(user_id)

        data_context = self.cache.get(data_key)
        return deepcopy(data_context) if data_context else default

    async def update_data(
        self, user_id: Union[int, str], data: Optional[Dict[Any, Any]] = None, **kwargs
    ) -> None:
        """Update data for a user.

        This method updates existing data with new values, similar to dict.update().
        Existing keys will be updated, and new keys will be added.

        Args:
            user_id (int | str): ID of the user
            data (dict[str, Any]): Data to update

        Example:
            >>> # Existing data: {"name": "John"}
            >>> await storage.update_data(user_id, {"age": 25})
            >>> # Result: {"name": "John", "age": 25}
        """
        if data is not None and not isinstance(data, dict):
            raise ValueError(f"'data' must be a dict, got {type(data)}")

        data_payload = {}
        if data:
            data_payload.update(data)
        if kwargs:
            data_payload.update(kwargs)

        if not data_payload:
            raise ValueError("No data passed.")

        data_key = self._get_data_key(user_id)

        data_context: dict = self.cache.get(data_key)
        if data_context is None:
            self.cache[data_key] = deepcopy(data_payload)
        else:
            data_context.update(deepcopy(data_payload))

    async def delete_data(
        self, user_id: Union[int, str], default: Optional[Any] = None
    ) -> Optional[Dict[Any, Any]]:
        """Clear and get all data for a user.

        Args:
            user_id (int | str): ID of the user
            default (Any, optional): Default value if data doesn't exist.
                Defaults to None.

        Returns:
            Dict | None: The deleted data or default value
        """
        data_key = self._get_data_key(user_id)
        return self.cache.pop(data_key, default)



from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, Optional, Union


class BaseAsyncStorage(ABC):
    """
    This is base class for Storage classes like MemoryStateStorage and RedisStateStorage
    """

    @abstractmethod
    async def set_state(self, key: Union[str, int], state: Union[str, Enum]) -> None:
        """
        use this method to set state for a key
        """
        raise NotImplementedError

    @abstractmethod
    async def get_state(
        self, key: Union[str, int], default: Optional[Any] = None
    ) -> Optional[str]:
        """
        use this method to get current state of a key

        return default if not found
        """
        raise NotImplementedError

    @abstractmethod
    async def delete_state(
        self, key: Union[str, int], default: Optional[Any] = None
    ) -> Optional[str]:
        """
        use this method to delete and get current state of a key

        return default if not found
        """
        raise NotImplementedError

    @abstractmethod
    async def set_data(
        self, key: Union[str, int], data: Optional[Dict[Any, Any]] = None, **kwargs
    ) -> None:
        """
        use this method to set data for a key
        """
        raise NotImplementedError

    @abstractmethod
    async def get_data(
        self, key: Union[str, int], default: Optional[Any] = None
    ) -> Dict[Any, Any]:
        """
        use this method to get current data of a key
        """
        raise NotImplementedError

    @abstractmethod
    async def update_data(
        self, key: int, data: Optional[Dict[Any, Any]] = None, **kwargs
    ) -> None:
        """
        use thid method to update current data of a key
        """
        raise NotImplementedError

    @abstractmethod
    async def delete_data(
        self, key: Union[str, int], default: Optional[Any] = None
    ) -> Optional[Dict[Any, Any]]:
        """
        use this method to clear and get current data of a key
        """
        raise NotImplementedError



"""
A Python library to handle steps in aiogram framework.
"""

__author__ = "Nasrollah Yusefi"
__version__ = "0.3.7"

__all__ = [
    "change_root_store",
    "register_next_step",
    "unregister_steps",
    "wait_for",
    "clear",
    "aiogram_dialect",
    "telebot_dialect",
    "telethon_dialect",
    "BaseStorage",
    "StateContext",
    "MemoryStateStorage",
    "FileStateStorage",
    "RedisStateStorage",
]

from .steps import MetaStore as MetaStore
from .steps import aiogram_dialect as aiogram_dialect
from .steps import change_root_store as change_root_store
from .steps import clear as clear
from .steps import register_next_step as register_next_step
from .steps import telebot_dialect as telebot_dialect
from .steps import telethon_dialect as telethon_dialect
from .steps import unregister_steps as unregister_steps
from .steps import wait_for as wait_for
from .storage import (
    BaseStorage,
    FileStateStorage,
    MemoryStateStorage,
    RedisStateStorage,
    StateContext,
)



from .base import BaseAsyncStorage
from .file import AsyncFileStateStorage
from .memory import AsyncMemoryStateStorage
from .redis import AsyncRedisStateStorage

__all__ = [
    "BaseAsyncStorage",
    "AsyncMemoryStateStorage",
    "AsyncRedisStateStorage",
    "AsyncFileStateStorage",
]



from .base import BaseStorage, StateContext
from .file import FileStateStorage
from .memory import MemoryStateStorage
from .redis import RedisStateStorage

__all__ = [
    "BaseStorage",
    "StateContext",
    "MemoryStateStorage",
    "RedisStateStorage",
    "FileStateStorage",
]



import os
import time
from enum import Enum
from typing import Any, Callable, Dict, Optional, Union

from qsave import QuickSave

from .base import BaseStorage, StateContext


class FileStateStorage(BaseStorage):
    """File-based storage implementation for managing bot states.

    This class provides a file-based storage solution for managing bot states
    and their associated data. Suitable for applications where persistence
    across sessions or restarts is required.

    Args:
        path (str | os.PathLike): Path to the file used for storing states and data.
    """

    def __init__(
        self, path: Union[str, os.PathLike], ex: Optional[float] = None, **kwargs
    ) -> None:
        """Initialize the file storage.

        Args:
            path (str | os.PathLike): File path to store states and data persistently.
            ex (float | None): expiry time for objects. pass in seconds.
        """
        self.cache = QuickSave(path=path, **kwargs)
        self.ex = ex

    def _get_key(self, user_id: Union[int, str]) -> str:
        """Generate Cache key for a user.

        Args:
            user_id (int | str): ID of the user

        Returns:
            str: Cache key
        """
        return f"state:{user_id}"

    def _get_data_key(self, user_id: Union[int, str]) -> str:
        """Generate Cache key for a user.

        Args:
            user_id (int | str): ID of the user

        Returns:
            str: Cache key
        """
        return f"data:{user_id}"

    def set_state(
        self,
        user_id: Union[int, str],
        state: Union[str, Enum],
        callback: Optional[Callable[..., Any]] = None,
        chat_id: Optional[Union[int, str]] = None,
        ex: Optional[float] = None,
    ) -> None:
        """Set the state for a user.

        Args:
            user_id (int | str): ID of the user
            state (str | Enum): State to set
            callback (Callable | None, optional): Callback function. Defaults to None.
            chat_id (int | str, optional): Chat ID. Defaults to None.
        """
        if chat_id is None:
            chat_id = user_id

        if isinstance(state, Enum):
            state = state.name

        callback_name = callback.__name__ if callback else None
        ex = ex or self.ex

        state_data = {
            "current_state": state,
            "chat_id": chat_id,
            "callback": callback_name,
        }
        ex = ex or self.ex
        if ex:
            state_data["expire"] = time.time() + ex
        state_key = self._get_key(user_id)

        with self.cache.session() as session:
            session[state_key] = state_data

    def get_state(
        self, user_id: Union[int, str], default: Optional[Any] = None
    ) -> Optional[StateContext]:
        """Get the state context for a user.

        Args:
            user_id (int | str): ID of the user
            default (Any, optional): Default value if state doesn't exist.
                Defaults to None.

        Returns:
            StateContext | None: The state context or default value
        """
        with self.cache.session(commit_on_expire=False) as session:
            data = session.get(self._get_key(user_id))

            if not data:
                return default
            expire = data.pop("expire", None)
            if expire and (expire < time.time()):
                session.pop(self._get_key(user_id))
                session.commit()
                return default

        return StateContext(**data)

    def delete_state(
        self, user_id: Union[int, str], default: Optional[Any] = None
    ) -> Optional[StateContext]:
        """Delete the state for a user.

        Args:
            user_id (int | str): ID of the user
            default (Any, optional): Default value if state doesn't exist.
                Defaults to None.

        Returns:
            StateContext | None: The deleted state context or default value
        """
        state_key = self._get_key(user_id)

        with self.cache.session() as session:
            data = session.pop(state_key)

        if not data:
            return default
        if data.get("expire") and (data.get("expire") < time.time()):
            return default

        return StateContext(**data)

    def set_data(
        self,
        user_id: Union[int, str],
        data: Optional[Dict[Any, Any]] = None,
        ex: Optional[float] = None,
        **kwargs,
    ) -> None:
        """Set data for a user.

        This method completely replaces any existing data.

        Args:
            user_id (int | str): ID of the user
            data (dict[str, Any]): Data to store
        """
        if data is not None and not isinstance(data, dict):
            raise ValueError(f"'data' must be a dict, got {type(data)}")

        data_payload = {}
        if data:
            data_payload.update(data)
        if kwargs:
            data_payload.update(kwargs)

        if not data_payload:
            raise ValueError("No data passed.")

        data_key = self._get_data_key(user_id)

        with self.cache.session() as session:
            ex = ex or self.ex
            new_data = {
                "data": data_payload,
            }
            if ex:
                new_data["ex"] = time.time() + ex
            session[data_key] = new_data

    def get_data(
        self, user_id: Union[int, str], default: Optional[Any] = None
    ) -> Optional[Dict[Any, Any]]:
        """Get data for a user.

        Args:
            user_id (int | str): ID of the user
            default (Any, optional): Default value if data doesn't exist.
                Defaults to None.

        Returns:
            dict[str, Any] | None: The stored data or None if not found
        """
        with self.cache.session(commit_on_expire=False) as session:
            data = session.get(self._get_data_key(user_id))

            if not data:
                return default
            if data.get("ex") and (data.get("ex") < time.time()):
                session.pop(self._get_data_key(user_id))
                session.commit()
                return default

        return data["data"]

    def update_data(
        self,
        user_id: Union[int, str],
        data: Optional[Dict[Any, Any]] = None,
        ex: Optional[float] = None,
        **kwargs,
    ) -> None:
        """Update data for a user.

        This method updates existing data with new values, similar to dict.update().
        Existing keys will be updated, and new keys will be added.

        Args:
            user_id (int | str): ID of the user
            data (dict[str, Any]): Data to update

        Example:
            >>> # Existing data: {"name": "John"}
            >>> storage.update_data(user_id, {"age": 25})
            >>> # Result: {"name": "John", "age": 25}
        """
        if data is not None and not isinstance(data, dict):
            raise ValueError(f"'data' must be a dict, got {type(data)}")

        data_payload = {}
        if data:
            data_payload.update(data)
        if kwargs:
            data_payload.update(kwargs)

        if not data_payload:
            raise ValueError("No data passed.")

        data_key = self._get_data_key(user_id)

        with self.cache.session() as session:
            current_data = session.get(data_key)

            if current_data:
                current_data["data"].update(data_payload)
                if ex:
                    current_data["ex"] = time.time() + ex
            else:
                ex = ex or self.ex
                new_data = {
                    "data": data_payload,
                }
                if ex:
                    new_data["ex"] = time.time() + ex
                session[data_key] = new_data

    def delete_data(
        self, user_id: Union[int, str], default: Optional[Any] = None
    ) -> Optional[Dict[Any, Any]]:
        """Clear and get all data for a user.

        Args:
            user_id (int | str): ID of the user
            default (Any, optional): Default value if data doesn't exist.
                Defaults to None.

        Returns:
            Dict | None: The deleted data or default value
        """
        data_key = self._get_data_key(user_id)
        with self.cache.session() as session:
            data = session.pop(data_key, default)

        if isinstance(data, dict):
            if data.get("expire") and (data.get("expire") < time.time()):
                return default

        return data



import asyncio
import typing

try:
    from aiogram import BaseMiddleware, types

    aiogram_installed = True
except ImportError:

    class BaseMiddleware:
        pass

    aiogram_installed = False

from ..functions import MetaStore, root


class Listen(BaseMiddleware):
    """
    aiogram middleware to listen for steps.

    supported handlers:
        - `MessageHandler`
        - `CallbackQueryHandler`
        - `ChatJoinRequestHandler`
        - `ChatMemberUpdatedHandler`
        - `EditedMessageHandler`

    Example::

        dp = Dispatcher()
        dp.message.outer_middleware(aiostep.Listen())
    """

    def __init__(self, store: typing.Optional[MetaStore] = None) -> None:
        if not aiogram_installed:
            raise ImportError(
                "aiogram package is not installed. install package: pip install aiogram"
            )
        self.store = store or root

    async def __call__(
        self,
        handler: typing.Callable[
            [types.TelegramObject, typing.Dict[str, typing.Any]],
            typing.Awaitable[typing.Any],
        ],
        event: types.TelegramObject,
        data: typing.Dict[str, typing.Any],
    ) -> typing.Any:
        fn = None

        try:
            fn = await self.store.pop_item(event.from_user.id)
        except (KeyError, AttributeError):
            try:
                chat_id = (
                    event.message.chat.id
                    if isinstance(event, types.CallbackQuery)
                    else event.chat.id
                )
                fn = await self.store.pop_item(chat_id)
            except (KeyError, AttributeError):
                pass

        if fn is not None:
            if isinstance(fn, asyncio.Future):
                fn.set_result(event)
                return
            await fn(event)
            return

        return await handler(event, data)



from .dialects import aiogram_dialect as aiogram_dialect
from .dialects import telebot_dialect as telebot_dialect
from .dialects import telethon_dialect as telethon_dialect
from .functions import MetaStore as MetaStore
from .functions import change_root_store as change_root_store
from .functions import clear as clear
from .functions import register_next_step as register_next_step
from .functions import unregister_steps as unregister_steps
from .functions import wait_for as wait_for
